extern crate bitvec;

use std::cell::{Cell, RefCell};
use std::collections::{HashMap, HashSet, VecDeque};
use std::fmt::{Error, Write};
use std::iter::once;

use self::bitvec::prelude::*;

use crate::*;

/*
 * List of big TODOs that aren't urgent:
 *
 * 1. Return `undef` when a PartitionExit data output isn't dominated by that
 *    datum's definition. PartitionExit always returns the complete set of data
 *    outputs that could ever be needed from a partition - this is because we
 *    don't want to introduce sum types into partition signatures or do funky
 *    pointer tricks. So, we could run into the following situation:
 *
 *                                 Block 1
 *                                /       \
 *                            Block 2   Block 3
 *                               |         |
 *                           Define v1  Define v2
 *                              /           \
 *               PartitionExit(v1,v2)    PartitionExit(v1, v2)
 *
 *    Let's assume that the PartitionExits branch to different partitions where
 *    one of them needs v1 and the other needs v2. Notice that both
 *    PartitionExits need to return both v1 and v2, since their signatures must
 *    be identical, even though for both, one of v1 or v2 doesn't dominate the
 *    PartitionExit. What *should* happen here is that each PartitionExit gets
 *    lowered to an LLVM `ret`, where the non-dominating output is set to
 *    `undef`. This works since in the original, un-partitioned, Hercules IR,
 *    defs must dominate uses, so we won't run into a situation where a returned
 *    `undef` value is actually read. What happens currently is that the
 *    generated LLVM will `ret` `%v1` and `%v2`, which LLVM won't compile (since
 *    the code wouldn't be in SSA form). This should get fixed when we start
 *    compiling more complicated codes.
 *
 * 2. Handle >= 3D fork-joins and array accesses. This isn't conceptually
 *    difficult, but generating the LLVM code to implement these is annoying.
 *
 * 3. Handle ABI properly when taking in / returning structs taking more than 16
 *    bytes. When a passed / returned struct takes more than 16 bytes, it needs
 *    to be passed around via pointers. This is one of many platform specific C
 *    ABI rules we need to handle to be properly called from Rust (that 16 byte
 *    rule is actually x86-64 specific). I'm honestly not sure how to handle
 *    this well. We avoid running into the manifestation of this problem for
 *    some samples by removing unneeded parameters / return values from
 *    partitions at the schedule IR level, which we should do anyway, but this
 *    isn't a complete solution.
 */

/*
 * The top level function to compile a schedule IR function into LLVM IR, for
 * execution on the CPU. We generate LLVM IR textually, since there are no good
 * LLVM bindings for Rust, and we are *not* writing any C++.
 */
pub fn cpu_compile<W: Write>(
    function: &SFunction,
    manifest: &PartitionManifest,
    w: &mut W,
) -> Result<(), Error> {
    // Calculate basic analyses over schedule IR.
    let virt_reg_to_inst_id = sched_virt_reg_to_inst_id(function);
    let dep_graph = sched_dependence_graph(function, &virt_reg_to_inst_id);
    let svalue_types = sched_svalue_types(function);
    let parallel_reduce_infos = sched_parallel_reduce_sections(function);

    // Calculate the names of each block. For blocks that are the top or bottom
    // blocks of sequential fork-joins, references outside the fork-join
    // actually need to refer to the header block. This is a bit complicated to
    // handle, and we use these names in several places, so pre-calculate the
    // block names. Intuitively, if we are "inside" a sequential fork-join,
    // references to the top or bottom blocks actually refer to those blocks,
    // while if we are "outside" the sequential fork-join, references to both
    // the top or bottom blocks actually refer to the loop header block.
    let mut block_names = HashMap::new();
    for (block_idx, block) in function.blocks.iter().enumerate() {
        for fork_join_id in parallel_reduce_infos
            .keys()
            .map(|id| Some(*id))
            .chain(once(None))
        {
            let block_id = BlockID::new(block_idx);
            let possible_parent = block.kind.try_fork_join_id();
            let mut walk = fork_join_id;

            // Check if the location in the key is "inside" the location of the
            // block.
            let is_inside = if let Some(parent) = possible_parent {
                loop {
                    if let Some(step) = walk {
                        if step == parent {
                            // If we see the block's location, then the key
                            // location is "inside".
                            break true;
                        } else {
                            // Walk the parent until we find something
                            // interesting.
                            walk = parallel_reduce_infos[&step].parent_fork_join_id;
                        }
                    } else {
                        // If we don't find the block, then the key location is
                        // "outside" the block's parallel-reduce.
                        break false;
                    }
                }
            } else {
                // Every location is "inside" the top level sequential section.
                true
            };

            // Check if the parent is a vectorized fork-join.
            let is_parent_vectorized = possible_parent
                // Check if the parent fork-join has a vector width.
                .map(|parent| parallel_reduce_infos[&parent].vector_width.is_some())
                // Sequential blocks are not vectorized.
                .unwrap_or(false);

            // If we are inside the block's fork-join or the block's fork-join
            // is vectorized, then refer to the blocks directly. Vectorized
            // fork-joins have the same LLVM IR control flow as the schedule IR
            // control flow.
            if is_inside || is_parent_vectorized {
                block_names.insert((block_id, fork_join_id), format!("bb_{}", block_idx));
            } else {
                block_names.insert(
                    (block_id, fork_join_id),
                    format!("fork_join_seq_header_{}", possible_parent.unwrap().idx()),
                );
            }
        }
    }

    // Create context for emitting LLVM IR.
    let ctx = CPUContext {
        function,
        manifest,
        block: Cell::new((0, &function.blocks[0])),

        virt_reg_to_inst_id,
        dep_graph,
        svalue_types,
        parallel_reduce_infos,

        block_names,

        vector_width: Cell::new(None),
        outside_def_used_in_vector: RefCell::new(HashSet::new()),
        vectors_from_parallel: RefCell::new(HashSet::new()),
        vector_reduce_associative_vars: RefCell::new(HashSet::new()),
        vector_reduce_cycle: Cell::new(false),
    };
    ctx.emit_function(w)?;

    Ok(())
}

/*
 * Top level structure to hold analysis info and cell-ed state.
 */
struct CPUContext<'a> {
    function: &'a SFunction,
    manifest: &'a PartitionManifest,
    block: Cell<(usize, &'a SBlock)>,

    // Basic analyses over schedule IR.
    virt_reg_to_inst_id: HashMap<usize, InstID>,
    dep_graph: HashMap<InstID, Vec<InstID>>,
    svalue_types: HashMap<SValue, SType>,
    parallel_reduce_infos: HashMap<ForkJoinID, ParallelReduceInfo>,

    // Calculate the names of each block up front. For blocks that are the top
    // or bottom blocks of sequential fork-joins, references outside the fork-
    // join actually need to refer to the header block. This is a bit
    // complicated to handle, and we use these names in several places, so pre-
    // calculate the block names. Intuitively, if we are "inside" a sequential
    // fork-join, references to the top or bottom blocks actually refer to those
    // blocks, while if we are "outside" the sequential fork-join, references to
    // both the top or bottom blocks actually refer to the loop header block.
    // Fully vectorized fork-joins are not considered sequential.
    block_names: HashMap<(BlockID, Option<ForkJoinID>), String>,

    // Track whether we are currently in a vectorized parallel section - this
    // affects how we lower types, for example.
    vector_width: Cell<Option<usize>>,
    // Track which virtual registers are defined outside the vectorized parallel
    // section and used within it.
    outside_def_used_in_vector: RefCell<HashSet<usize>>,
    // Track which virtual registers are defined in the vectorized parallel
    // section and used in the vectorized reduce section.
    vectors_from_parallel: RefCell<HashSet<usize>>,
    // Track which reduction variables (store their virtual register and
    // variable number) are associative in the vectorized reduce section.
    vector_reduce_associative_vars: RefCell<HashSet<(usize, usize)>>,
    // track whether there are any non-associative reduction variables in a
    // vectorized reduce section (which corresponds to whether we need to
    // generate explicit control flow or not).
    vector_reduce_cycle: Cell<bool>,
}

impl<'a> CPUContext<'a> {
    fn emit_function<W: Write>(&self, w: &mut W) -> Result<(), Error> {
        // Emit the partition function signature.
        write!(w, "define ")?;
        if self.function.return_types.len() == 1 {
            self.emit_type(&self.function.return_types[0], w)?;
        } else {
            // Functions with multiple return values return said values in a
            // struct.
            self.emit_type(
                &SType::Product(self.function.return_types.clone().into()),
                w,
            )?;
        }
        write!(w, " @{}(", self.manifest.name)?;
        (0..self.function.param_types.len())
            .map(|param_idx| Some(SValue::VirtualRegister(param_idx)))
            .intersperse(None)
            .map(|token| -> Result<(), Error> {
                match token {
                    Some(param_svalue) => {
                        self.emit_svalue(&param_svalue, true, w)?;
                    }
                    None => write!(w, ", ")?,
                }
                Ok(())
            })
            .collect::<Result<(), Error>>()?;
        // Technically, this may fail if for some reason there's a parallel
        // launch partition with no parameters. Blame LLVM for being
        // unnecessarily strict about commas of all things...
        for parallel_launch_dim in 0..self.manifest.device.num_parallel_launch_dims() {
            write!(w, ", i64 %parallel_launch_{}_low", parallel_launch_dim)?;
            write!(w, ", i64 %parallel_launch_{}_len", parallel_launch_dim)?;
        }
        write!(w, ") {{\n",)?;

        // Emit the function body. Emit each block, one at a time.
        for (block_idx, block) in self.function.blocks.iter().enumerate() {
            self.block.set((block_idx, block));

            // For "tops" of sequential fork-joins, we emit a special top block
            // to be the loop header for the fork-join loop.
            if let Some(fork_join_id) = block.kind.try_parallel()
                && self.parallel_reduce_infos[&fork_join_id]
                    .top_parallel_block
                    .idx()
                    == block_idx
                && self.parallel_reduce_infos[&fork_join_id]
                    .vector_width
                    .is_none()
            {
                self.emit_fork_join_seq_header(fork_join_id, block_idx, w)?;
            }

            // Emit the header for the block.
            write!(
                w,
                "{}:\n",
                &self.block_names[&(BlockID::new(block_idx), block.kind.try_fork_join_id())]
            )?;

            // If this block is in a vectorized parallel section, set up the
            // context for vector code generation.
            if let Some(fork_join_id) = block.kind.try_parallel()
                && let Some(width) = self.parallel_reduce_infos[&fork_join_id].vector_width
            {
                self.setup_vectorized_parallel_block(width, w)?;
            }

            // If this block is in a vectorized reduce section, set up either a
            // post-parallel reduction loop or a vector reduction, depending on
            // whether there's an associative schedule on each reduction
            // variable.
            if let Some(fork_join_id) = block.kind.try_reduce()
                && let Some(width) = self.parallel_reduce_infos[&fork_join_id].vector_width
            {
                self.setup_vectorized_reduce_block(fork_join_id, width, w)?;
            }

            // For each basic block, emit instructions in that block. Emit using
            // a worklist over the dependency graph.
            let mut emitted = bitvec![u8, Lsb0; 0; block.insts.len()];
            let mut worklist = VecDeque::from((0..block.insts.len()).collect::<Vec<_>>());
            while let Some(inst_idx) = worklist.pop_front() {
                let inst_id = InstID::new(block_idx, inst_idx);
                let dependencies = &self.dep_graph[&inst_id];
                let all_uses_emitted = dependencies
                    .into_iter()
                    // Check that all used instructions in this block...
                    .filter(|inst_id| inst_id.idx_0() == block_idx)
                    // were already emitted.
                    .all(|inst_id| emitted[inst_id.idx_1()]);
                // Phis don't need to wait for all of their uses to be emitted.
                if block.insts[inst_idx].is_phi() || all_uses_emitted {
                    self.emit_inst(
                        block.virt_regs[inst_id.idx_1()].0,
                        &block.insts[inst_idx],
                        block.kind.try_fork_join_id(),
                        w,
                    )?;
                    emitted.set(inst_id.idx_1(), true);
                } else {
                    worklist.push_back(inst_idx);
                }
            }

            self.reset_cells();
        }
        write!(w, "}}\n",)?;

        Ok(())
    }

    fn emit_type<W: Write>(&self, stype: &SType, w: &mut W) -> Result<(), Error> {
        if let Some(width) = self.vector_width.get() {
            write!(w, "<{} x ", width)?;
        }

        match stype {
            SType::Boolean => write!(w, "i1")?,
            SType::Integer8 | SType::UnsignedInteger8 => write!(w, "i8")?,
            SType::Integer16 | SType::UnsignedInteger16 => write!(w, "i16")?,
            SType::Integer32 | SType::UnsignedInteger32 => write!(w, "i32")?,
            SType::Integer64 | SType::UnsignedInteger64 => write!(w, "i64")?,
            SType::Float32 => write!(w, "float")?,
            SType::Float64 => write!(w, "double")?,
            SType::Product(fields) => {
                write!(w, "{{")?;
                fields
                    .into_iter()
                    .map(Some)
                    .intersperse(None)
                    .map(|token| -> Result<(), Error> {
                        match token {
                            Some(field_ty) => self.emit_type(field_ty, w)?,
                            None => write!(w, ", ")?,
                        }
                        Ok(())
                    })
                    .collect::<Result<(), Error>>()?;
                write!(w, "}}")?;
            }
            SType::ArrayRef(_) => write!(w, "ptr")?,
        }

        if self.vector_width.get().is_some() {
            write!(w, ">")?;
        }

        Ok(())
    }

    fn emit_constant<W: Write>(&self, sconstant: &SConstant, w: &mut W) -> Result<(), Error> {
        match sconstant {
            SConstant::Boolean(val) => write!(w, "{}", val)?,
            SConstant::Integer8(val) => write!(w, "{}", val)?,
            SConstant::Integer16(val) => write!(w, "{}", val)?,
            SConstant::Integer32(val) => write!(w, "{}", val)?,
            SConstant::Integer64(val) => write!(w, "{}", val)?,
            SConstant::UnsignedInteger8(val) => write!(w, "{}", val)?,
            SConstant::UnsignedInteger16(val) => write!(w, "{}", val)?,
            SConstant::UnsignedInteger32(val) => write!(w, "{}", val)?,
            SConstant::UnsignedInteger64(val) => write!(w, "{}", val)?,
            SConstant::Float32(val) => {
                if val.fract() == 0.0 {
                    write!(w, "{}.0", val)?
                } else {
                    write!(w, "{}", val)?
                }
            }
            SConstant::Float64(val) => {
                if val.fract() == 0.0 {
                    write!(w, "{}.0", val)?
                } else {
                    write!(w, "{}", val)?
                }
            }
            SConstant::Product(fields) => {
                write!(w, "{{")?;
                fields
                    .into_iter()
                    .map(Some)
                    .intersperse(None)
                    .map(|token| -> Result<(), Error> {
                        match token {
                            Some(field_cons) => {
                                self.emit_type(&field_cons.get_type(), w)?;
                                write!(w, " ")?;
                                self.emit_constant(field_cons, w)?;
                            }
                            None => write!(w, ", ")?,
                        }
                        Ok(())
                    })
                    .collect::<Result<(), Error>>()?;
                write!(w, "}}")?;
            }
        }

        Ok(())
    }

    fn emit_svalue<W: Write>(&self, svalue: &SValue, add_ty: bool, w: &mut W) -> Result<(), Error> {
        if add_ty {
            self.emit_type(&self.svalue_types[svalue], w)?;
            write!(w, " ")?;
        }
        if self.vector_width.get().is_some()
            && svalue
                .try_virt_reg()
                .map(|virt_reg| self.outside_def_used_in_vector.borrow().contains(&virt_reg))
                .unwrap_or(false)
        {
            match svalue {
                SValue::VirtualRegister(virt_reg) => {
                    write!(w, "%vec_{}_v{}", self.block.get().0, virt_reg)?
                }
                SValue::Constant(_) => todo!(),
            }
        } else if svalue
            .try_virt_reg()
            .map(|virt_reg| self.vectors_from_parallel.borrow().contains(&virt_reg))
            .unwrap_or(false)
        {
            match svalue {
                SValue::VirtualRegister(virt_reg) => write!(w, "%extract_v{}", virt_reg)?,
                SValue::Constant(_) => todo!(),
            }
        } else {
            match svalue {
                SValue::VirtualRegister(virt_reg) => write!(w, "%v{}", virt_reg)?,
                SValue::Constant(cons) => self.emit_constant(cons, w)?,
            }
        }
        Ok(())
    }

    fn emit_inst<W: Write>(
        &self,
        virt_reg: usize,
        inst: &SInst,
        location: Option<ForkJoinID>,
        w: &mut W,
    ) -> Result<(), Error> {
        // Helper to emit the initial assignment to the destination virtual
        // register, when applicable.
        let self_svalue = SValue::VirtualRegister(virt_reg);
        let emit_assign = |w: &mut W| -> Result<(), Error> { write!(w, "%v{} = ", virt_reg) };

        write!(w, "  ")?;
        match inst {
            SInst::Phi { inputs } => {
                emit_assign(w)?;
                write!(w, "phi ")?;
                self.emit_type(&self.svalue_types[&self_svalue], w)?;
                write!(w, " ")?;
                inputs
                    .into_iter()
                    .map(Some)
                    .intersperse(None)
                    .map(|token| match token {
                        Some((pred_block_id, svalue)) => {
                            write!(w, "[ ")?;
                            self.emit_svalue(svalue, false, w)?;
                            write!(w, ", %{} ]", &self.block_names[&(*pred_block_id, location)])?;
                            Ok(())
                        }
                        None => write!(w, ", "),
                    })
                    .collect::<Result<(), Error>>()?;
            }
            SInst::ThreadID {
                dimension,
                fork_join,
            } => {
                emit_assign(w)?;
                if let Some(width) = self.vector_width.get() {
                    write!(w, "add <{} x i64> <", width)?;
                    for idx in 0..width {
                        if idx != 0 {
                            write!(w, ", ")?;
                        }
                        write!(w, "i64 {}", idx)?;
                    }
                    write!(w, ">, zeroinitializer")?;
                } else {
                    write!(w, "add i64 0, %thread_id_{}_{}", fork_join.idx(), dimension)?;
                }
            }
            SInst::ReductionVariable { number } => {
                write!(w, "; Already emitted reduction variable #{number}.")?;
            }
            SInst::Jump {
                target,
                parallel_entry: _,
                reduce_exit,
            } => {
                if reduce_exit.is_some() && self.vector_reduce_cycle.get() {
                    // If we're closing a non-vectorized reduction for a
                    // vectorized parallel, jump back to the beginning of the
                    // reduction, not the beginning of the parallel section.
                    let self_block_idx = self.block.get().0;
                    write!(
                        w,
                        "br label %{}",
                        &self.block_names[&(BlockID::new(self_block_idx), location)],
                    )?;
                } else if reduce_exit.is_some() && self.vectors_from_parallel.borrow().is_empty() {
                    // If we're closing a reduction and the parallel-reduce is
                    // not vectorized, we need to branch back to the beginning
                    // of the parallel-reduce.
                    write!(
                        w,
                        "br label %fork_join_seq_header_{}",
                        location.unwrap().idx(),
                    )?;
                } else {
                    // If this is a normal jump (or is closing a reduction and
                    // is vectorized, along with the parallel section), then
                    // branch to the successor as expected.
                    write!(w, "br label %{}", &self.block_names[&(*target, location)])?;
                }
            }
            SInst::Branch {
                cond,
                false_target,
                true_target,
            } => {
                // Branches aren't involved in any parallel-reduce shenanigans,
                // so lowering them is straightforward.
                write!(w, "br ")?;
                self.emit_svalue(cond, true, w)?;
                write!(
                    w,
                    ", label %{}, label %{}",
                    &self.block_names[&(*true_target, location)],
                    &self.block_names[&(*false_target, location)],
                )?;
            }
            SInst::PartitionExit { data_outputs } => {
                if data_outputs.len() == 0 {
                    write!(w, "ret {{}} zeroinitializer")?;
                } else if data_outputs.len() == 1 {
                    write!(w, "ret ")?;
                    self.emit_svalue(&data_outputs[0], true, w)?;
                } else {
                    let ret_ty = SType::Product(
                        data_outputs
                            .iter()
                            .map(|svalue| self.svalue_types[svalue].clone())
                            .collect(),
                    );
                    write!(w, "%v{}_0 = insertvalue ", virt_reg)?;
                    self.emit_type(&ret_ty, w)?;
                    write!(w, " undef, ")?;
                    self.emit_svalue(&data_outputs[0], true, w)?;
                    write!(w, ", 0\n")?;
                    for idx in 1..data_outputs.len() {
                        write!(w, "  %v{}_{} = insertvalue ", virt_reg, idx)?;
                        self.emit_type(&ret_ty, w)?;
                        write!(w, " %v{}_{}, ", virt_reg, idx - 1)?;
                        self.emit_svalue(&data_outputs[idx], true, w)?;
                        write!(w, ", {}\n", idx)?;
                    }
                    write!(w, "  ret ")?;
                    self.emit_type(&ret_ty, w)?;
                    write!(w, " %v{}_{}", virt_reg, data_outputs.len() - 1)?;
                }
            }
            SInst::Return { value } => {
                write!(w, "ret ")?;
                self.emit_svalue(value, true, w)?;
            }
            SInst::Unary { input, op } => {
                emit_assign(w)?;
                match op {
                    SUnaryOperator::Not => {
                        write!(w, "xor ")?;
                        self.emit_svalue(input, true, w)?;
                        write!(w, ", -1")?;
                    }
                    SUnaryOperator::Neg => {
                        if self.svalue_types[input].is_float() {
                            write!(w, "fneg ")?;
                            self.emit_svalue(input, true, w)?;
                        } else {
                            write!(w, "mul ")?;
                            self.emit_svalue(input, true, w)?;
                            write!(w, ", -1")?;
                        }
                    }
                    SUnaryOperator::Cast(_) => todo!(),
                }
            }
            SInst::Binary { left, right, op } => {
                // If we're in a vectorized reduce block and this binary
                // operation is reducing over an associative reduction variable,
                // then we need to emit a LLVM vector reduce intrinsic.
                // Otherwise lower into a normal LLVM binary op.
                let try_associative_reduction = |sval: &SValue| {
                    sval.try_virt_reg()
                        .map(|virt_reg| {
                            self.vector_reduce_associative_vars
                                .borrow()
                                .iter()
                                .filter(|(red_virt_reg, _)| *red_virt_reg == virt_reg)
                                .map(|(red_virt_reg, red_num)| (*red_virt_reg, *red_num))
                                .next()
                        })
                        .flatten()
                };
                if let Some((red_virt_reg, red_num)) =
                    try_associative_reduction(left).or(try_associative_reduction(right))
                {
                    let left_virt_reg = left
                        .try_virt_reg()
                        .expect("PANIC: Associative reduction can't involve constants.");
                    let right_virt_reg = right
                        .try_virt_reg()
                        .expect("PANIC: Associative reduction can't involve constants.");
                    let vector_virt_reg = if left_virt_reg != red_virt_reg {
                        left_virt_reg
                    } else if right_virt_reg != red_virt_reg {
                        right_virt_reg
                    } else {
                        panic!("PANIC: Associative reduction can't use the reduction variable more than once.");
                    };
                    let info = &self.parallel_reduce_infos[&location.unwrap()];
                    write!(w, "%v{} = call reassoc ", red_virt_reg)?;
                    self.emit_type(&self.svalue_types[&self_svalue], w)?;
                    let op = op.get_llvm_op(&self.svalue_types[left]);
                    write!(w, " @llvm.vector.reduce.{}", op)?;
                    let width = info.vector_width.unwrap();
                    self.emit_reduce_suffix(width, &self.svalue_types[&self_svalue], w)?;
                    write!(w, "(")?;
                    self.emit_svalue(&info.reduce_inits[red_num], true, w)?;
                    write!(w, ", ")?;
                    self.vector_width.set(Some(width));
                    let old_vectors_from_parallel = self.vectors_from_parallel.take();
                    self.emit_svalue(&SValue::VirtualRegister(vector_virt_reg), true, w)?;
                    self.vector_width.set(None);
                    self.vectors_from_parallel
                        .replace(old_vectors_from_parallel);
                    write!(w, ")")?;
                } else {
                    emit_assign(w)?;
                    let op = op.get_llvm_op(&self.svalue_types[left]);
                    write!(w, "{} ", op)?;
                    self.emit_svalue(left, true, w)?;
                    write!(w, ", ")?;
                    self.emit_svalue(right, false, w)?;
                }
            }
            SInst::Ternary {
                first,
                second,
                third,
                op,
            } => {
                emit_assign(w)?;
                match op {
                    STernaryOperator::Select => {
                        write!(w, "select ")?;
                        self.emit_svalue(first, true, w)?;
                        write!(w, ", ")?;
                        self.emit_svalue(second, true, w)?;
                        write!(w, ", ")?;
                        self.emit_svalue(third, true, w)?;
                    }
                }
            }
            SInst::ArrayLoad {
                array,
                position,
                bounds,
            } => {
                self.emit_linear_index_calc(virt_reg, position, bounds, w)?;
                write!(w, "%load_ptr_{} = getelementptr ", virt_reg)?;
                let old_width = self.vector_width.take();
                self.emit_type(&self.svalue_types[&self_svalue], w)?;
                self.vector_width.set(old_width);
                write!(w, ", ")?;
                self.emit_svalue(array, true, w)?;
                write!(w, ", ")?;
                self.emit_type(&self.svalue_types[&position[0]], w)?;
                write!(w, " %calc_linear_idx_{}\n  ", virt_reg)?;
                emit_assign(w)?;
                if let Some(width) = self.vector_width.get() {
                    write!(w, "call ")?;
                    self.emit_type(&self.svalue_types[&self_svalue], w)?;
                    write!(w, " @llvm.masked.gather")?;
                    self.emit_gather_scatter_suffix(width, &self.svalue_types[&self_svalue], w)?;
                    write!(w, "(")?;
                    self.emit_type(&self.svalue_types[array], w)?;
                    write!(w, " %load_ptr_{}, i32 8, <{} x i1> <", virt_reg, width)?;
                    for idx in 0..width {
                        if idx != 0 {
                            write!(w, ", ")?;
                        }
                        write!(w, "i1 true")?;
                    }
                    write!(w, ">, ")?;
                    self.emit_type(&self.svalue_types[&self_svalue], w)?;
                    write!(w, " undef)")?;
                } else {
                    write!(w, "load ")?;
                    self.emit_type(&self.svalue_types[&self_svalue], w)?;
                    write!(w, ", ptr %load_ptr_{}", virt_reg)?;
                }
            }
            SInst::ArrayStore {
                array,
                value,
                position,
                bounds,
            } => {
                self.emit_linear_index_calc(virt_reg, position, bounds, w)?;
                write!(w, "%store_ptr_{} = getelementptr ", virt_reg)?;
                let old_width = self.vector_width.take();
                self.emit_type(&self.svalue_types[value], w)?;
                self.vector_width.set(old_width);
                write!(w, ", ")?;
                self.emit_svalue(array, true, w)?;
                write!(w, ", ")?;
                self.emit_type(&self.svalue_types[&position[0]], w)?;
                write!(w, " %calc_linear_idx_{}\n  ", virt_reg)?;
                if let Some(width) = self.vector_width.get() {
                    write!(w, "call ")?;
                    self.emit_type(&self.svalue_types[&self_svalue], w)?;
                    write!(w, " @llvm.masked.scatter")?;
                    self.emit_gather_scatter_suffix(width, &self.svalue_types[&self_svalue], w)?;
                    write!(w, "(")?;
                    self.emit_svalue(array, true, w)?;
                    write!(w, ", ")?;
                    self.emit_type(&self.svalue_types[array], w)?;
                    write!(w, " %store_ptr_{}, i32 8, <{} x i1> <", virt_reg, width)?;
                    for idx in 0..width {
                        if idx != 0 {
                            write!(w, ", ")?;
                        }
                        write!(w, "i1 true")?;
                    }
                    write!(w, ">)")?;
                } else {
                    write!(w, "store ")?;
                    self.emit_svalue(value, true, w)?;
                    write!(w, ", ptr %store_ptr_{}", virt_reg)?;
                }
            }
            _ => {}
        }
        write!(w, "\n")?;

        Ok(())
    }

    /*
     * Implement the index math to convert a multi-dimensional position to a
     * linear position inside an array.
     */
    fn emit_linear_index_calc<W: Write>(
        &self,
        virt_reg: usize,
        position: &[SValue],
        bounds: &[SValue],
        w: &mut W,
    ) -> Result<(), Error> {
        assert_eq!(position.len(), bounds.len());

        if position.len() == 1 {
            write!(w, "%calc_linear_idx_{} = add ", virt_reg)?;
            self.emit_svalue(&position[0], true, w)?;
            write!(w, ", zeroinitializer\n  ")?;
        } else if position.len() == 2 {
            write!(w, "%calc_linear_idx_{}_0 = mul ", virt_reg)?;
            self.emit_svalue(&position[0], true, w)?;
            write!(w, ", ")?;
            self.emit_svalue(&bounds[1], false, w)?;
            write!(w, "\n  %calc_linear_idx_{} = add ", virt_reg)?;
            self.emit_svalue(&position[1], true, w)?;
            write!(w, ", %calc_linear_idx_{}_0", virt_reg)?;
            write!(w, "\n  ")?;
        } else {
            todo!("TODO: Handle the 3 or more dimensional array case.")
        }

        Ok(())
    }

    /*
     * LLVM intrinsics are a pain to emit textually...
     */
    fn intrinsic_type_str(elem_ty: &SType) -> &'static str {
        // We can't just use our previous routines for emitting types, because
        // only inside intrinsics does LLVM use "f32" and "f64" properly!
        match elem_ty {
            SType::Boolean => "i1",
            SType::Integer8 | SType::UnsignedInteger8 => "i8",
            SType::Integer16 | SType::UnsignedInteger16 => "i16",
            SType::Integer32 | SType::UnsignedInteger32 => "i32",
            SType::Integer64 | SType::UnsignedInteger64 => "i64",
            SType::Float32 => "f32",
            SType::Float64 => "f64",
            _ => panic!(),
        }
    }

    fn emit_reduce_suffix<W: Write>(
        &self,
        width: usize,
        elem_ty: &SType,
        w: &mut W,
    ) -> Result<(), Error> {
        write!(w, ".v{}{}", width, Self::intrinsic_type_str(elem_ty))?;
        Ok(())
    }

    fn emit_gather_scatter_suffix<W: Write>(
        &self,
        width: usize,
        elem_ty: &SType,
        w: &mut W,
    ) -> Result<(), Error> {
        write!(
            w,
            ".v{}{}.v{}p0",
            width,
            Self::intrinsic_type_str(elem_ty),
            width
        )?;
        Ok(())
    }

    /*
     * Emit the loop header implementing a sequential fork-join. For historical
     * reasons, "sequential" fork-joins are just fork-joins that are lowered to
     * LLVM level loops. This includes fork-joins that end up getting
     * parallelized across threads via low/high bounds.
     */
    fn emit_fork_join_seq_header<W: Write>(
        &self,
        fork_join_id: ForkJoinID,
        block_idx: usize,
        w: &mut W,
    ) -> Result<(), Error> {
        let info = &self.parallel_reduce_infos[&fork_join_id];
        let entry_name = &self.block_names[&(info.predecessor, Some(fork_join_id))];
        let loop_name = &self.block_names[&(info.reduce_block, Some(fork_join_id))];
        let parallel_launch = self.manifest.device.num_parallel_launch_dims() > 0 && info.top_level;

        // Start the header of the loop.
        write!(w, "fork_join_seq_header_{}:\n", fork_join_id.idx())?;

        // Emit the phis for the linear loop index variable and the reduction
        // variables.
        write!(
            w,
            "  %linear_{} = phi i64 [ 0, %{} ], [ %linear_{}_inc, %{} ]\n",
            block_idx, entry_name, block_idx, loop_name,
        )?;
        for (var_num, virt_reg) in info.reduction_variables.iter() {
            write!(w, "  %v{} = phi ", virt_reg)?;
            self.emit_type(&self.svalue_types[&SValue::VirtualRegister(*virt_reg)], w)?;
            write!(w, " [ ")?;
            self.emit_svalue(&info.reduce_inits[*var_num], false, w)?;
            write!(w, ", %{} ], [ ", entry_name)?;
            self.emit_svalue(&info.reduce_reducts[*var_num], false, w)?;
            write!(w, ", %{} ]\n", loop_name)?;
        }

        // Calculate the loop bounds.
        if info.thread_counts.len() == 1 {
            write!(w, "  %bound_{} = add i64 0, ", block_idx)?;
            if parallel_launch {
                write!(w, "%parallel_launch_0_len")?;
            } else {
                self.emit_svalue(&info.thread_counts[0], false, w)?;
            }
            write!(w, "\n")?;
        } else if info.thread_counts.len() == 2 {
            write!(w, "  %bound_{} = mul ", block_idx)?;
            if parallel_launch {
                write!(w, "i64 %parallel_launch_0_len, %parallel_launch_1_len")?;
            } else {
                self.emit_svalue(&info.thread_counts[0], true, w)?;
                write!(w, ", ")?;
                self.emit_svalue(&info.thread_counts[1], false, w)?;
            }
            write!(w, "\n")?;
        } else {
            todo!("TODO: Handle the 3 or more dimensional fork-join case.")
        }

        // Calculate the multi-dimensional thread indices.
        if info.thread_counts.len() == 1 && parallel_launch {
            write!(
                w,
                "  %thread_id_{}_0 = add i64 %parallel_launch_0_low, %linear_{}\n",
                fork_join_id.idx(),
                block_idx
            )?;
        } else if info.thread_counts.len() == 1 {
            write!(
                w,
                "  %thread_id_{}_0 = add i64 0, %linear_{}\n",
                fork_join_id.idx(),
                block_idx
            )?;
        } else if info.thread_counts.len() == 2 && parallel_launch {
            write!(
                w,
                "  %unshifted_id_{}_0 = udiv i64 %linear_{}, %parallel_launch_1_len\n",
                fork_join_id.idx(),
                block_idx
            )?;
            write!(
                w,
                "  %unshifted_id_{}_1 = urem i64 %linear_{}, %parallel_launch_1_len\n",
                fork_join_id.idx(),
                block_idx
            )?;
            write!(
                w,
                "  %thread_id_{}_0 = add i64 %unshifted_id_{}_0, %parallel_launch_0_low\n",
                fork_join_id.idx(),
                fork_join_id.idx(),
            )?;
            write!(
                w,
                "  %thread_id_{}_1 = add i64 %unshifted_id_{}_1, %parallel_launch_1_low\n",
                fork_join_id.idx(),
                fork_join_id.idx(),
            )?;
        } else if info.thread_counts.len() == 2 {
            write!(
                w,
                "  %thread_id_{}_0 = udiv i64 %linear_{}, ",
                fork_join_id.idx(),
                block_idx
            )?;
            self.emit_svalue(&info.thread_counts[1], false, w)?;
            write!(w, "\n")?;
            write!(
                w,
                "  %thread_id_{}_1 = urem i64 %linear_{}, ",
                fork_join_id.idx(),
                block_idx
            )?;
            self.emit_svalue(&info.thread_counts[1], false, w)?;
            write!(w, "\n")?;
        } else {
            todo!("TODO: Handle the 3 or more dimensional fork-join case.")
        }

        // Increment the linear index.
        write!(
            w,
            "  %linear_{}_inc = add i64 %linear_{}, 1\n",
            block_idx, block_idx
        )?;

        // Emit the branch.
        write!(
            w,
            "  %cond_{} = icmp ult i64 %linear_{}, %bound_{}\n",
            block_idx, block_idx, block_idx
        )?;
        let top_name = &self.block_names[&(BlockID::new(block_idx), Some(fork_join_id))];
        let succ_name = &self.block_names[&(info.successor, Some(fork_join_id))];
        write!(
            w,
            "  br i1 %cond_{}, label %{}, label %{}\n",
            block_idx, top_name, succ_name
        )?;

        Ok(())
    }

    /*
     * Calculate and emit block-level info for vectorized parallel blocks.
     */
    fn setup_vectorized_parallel_block<W: Write>(
        &self,
        width: usize,
        w: &mut W,
    ) -> Result<(), Error> {
        let (block_idx, block) = self.block.get();

        // Get the uses of virtual registers defined outside the
        // vectorized region.
        let mut outside_def_used_in_vector = HashSet::new();
        for inst in block.insts.iter() {
            for virt_reg in sched_get_uses(inst).filter_map(|svalue| svalue.try_virt_reg()) {
                let outside = match self.virt_reg_to_inst_id.get(&virt_reg) {
                    Some(use_inst_id) => {
                        block.kind != self.function.blocks[use_inst_id.idx_0()].kind
                    }
                    // Parameters are always defined outside the vectorized
                    // region as scalars.
                    None => true,
                };
                if outside {
                    outside_def_used_in_vector.insert(virt_reg);
                }
            }
        }

        // Broadcast scalar values into vector values. The vector
        // register produced needs to be indexed in name by the block
        // index. This is because we may end up using the same value in
        // multiple vectorized blocks, and we can't have those
        // vectorized scalars have the same name.
        for outside_virt_reg in outside_def_used_in_vector.iter() {
            write!(
                w,
                "  %vec1_{}_v{} = insertelement <1 x ",
                block_idx, outside_virt_reg
            )?;
            let elem_ty = &self.svalue_types[&SValue::VirtualRegister(*outside_virt_reg)];
            self.emit_type(elem_ty, w)?;
            write!(w, "> undef, ")?;
            self.emit_type(elem_ty, w)?;
            write!(w, " %v{}, i32 0\n", outside_virt_reg)?;
            write!(
                w,
                "  %vec_{}_v{} = shufflevector <1 x ",
                block_idx, outside_virt_reg
            )?;
            self.emit_type(elem_ty, w)?;
            write!(w, "> %vec1_{}_v{}, <1 x ", block_idx, outside_virt_reg)?;
            self.emit_type(elem_ty, w)?;
            write!(w, "> undef, <{} x i32> zeroinitializer\n", width)?;
        }

        // Set the cell values in the context.
        self.vector_width.set(Some(width));
        self.outside_def_used_in_vector
            .replace(outside_def_used_in_vector);

        Ok(())
    }

    /*
     * Calculate and emit block-level info for vectorized reduce blocks.
     */
    fn setup_vectorized_reduce_block<W: Write>(
        &self,
        fork_join_id: ForkJoinID,
        width: usize,
        w: &mut W,
    ) -> Result<(), Error> {
        let (block_idx, block) = self.block.get();

        // Get uses of vector values defined in the parallel region.
        let mut vectors_from_parallel = HashSet::new();
        for inst in block.insts.iter() {
            for virt_reg in sched_get_uses(inst).filter_map(|svalue| svalue.try_virt_reg()) {
                if let Some(inst_id) = self.virt_reg_to_inst_id.get(&virt_reg)
                    && self.function.blocks[inst_id.idx_0()].kind
                        == SBlockKind::Parallel(fork_join_id)
                {
                    vectors_from_parallel.insert(virt_reg);
                }
            }
        }

        // Each reduction may be representable by an LLVM reduction intrinsic.
        // If every reduction in this reduce block is, then we don't need to
        // generate an explicit loop. If any one reduction isn't representable
        // as a single intrinsic, then we need to generate an explicit loop. The
        // explicit loop calculates the reduction for all reductions that can't
        // be represented by intrinsics, while intrinsics are still used to
        // calculate reductions that can be represented by them. Currently, the
        // "associative" schedule captures this info per reduction variable.
        let all_intrinsic_representable = block
            .insts
            .iter()
            .enumerate()
            .filter(|(_, inst)| inst.is_reduction_variable())
            .all(|(inst_idx, _)| block.schedules[&inst_idx].contains(&SSchedule::Associative));
        if !all_intrinsic_representable {
            let info = &self.parallel_reduce_infos[&fork_join_id];
            let entry_name = &self.block_names[&(info.bottom_parallel_block, Some(fork_join_id))];
            let self_name = &self.block_names[&(info.reduce_block, Some(fork_join_id))];
            let succ_name = &self.block_names[&(info.successor, Some(fork_join_id))];

            // Emit a loop header for the reduce.
            write!(
                w,
                "  %linear_{} = phi i64 [ 0, %{} ], [ %linear_{}_inc, %{}_reduce_body ]\n",
                block_idx, entry_name, block_idx, self_name,
            )?;
            // Emit phis for reduction variables here, since they need to be
            // above everything emitted below.
            for (var_num, virt_reg) in info.reduction_variables.iter() {
                // Only emit phis for reduction variables that aren't
                // implemented in intrinsics.
                if !block.schedules[&self.virt_reg_to_inst_id[virt_reg].idx_1()]
                    .contains(&SSchedule::Associative)
                {
                    write!(w, "  %v{} = phi ", virt_reg)?;
                    self.emit_type(&self.svalue_types[&SValue::VirtualRegister(*virt_reg)], w)?;
                    write!(w, " [ ")?;
                    self.emit_svalue(&info.reduce_inits[*var_num], false, w)?;
                    write!(w, ", %{} ], [ ", entry_name)?;
                    self.emit_svalue(&info.reduce_reducts[*var_num], false, w)?;
                    write!(w, ", %{}_reduce_body ]\n", self_name)?;
                }
            }
            write!(
                w,
                "  %linear_{}_inc = add i64 %linear_{}, 1\n",
                block_idx, block_idx
            )?;
            // The loop bound is the constant vector width.
            write!(
                w,
                "  %cond_{} = icmp ult i64 %linear_{}, {}\n",
                block_idx, block_idx, width
            )?;
            // Branch to the reduce loop body.
            write!(
                w,
                "  br i1 %cond_{}, label %{}_reduce_body, label %{}\n",
                block_idx, self_name, succ_name
            )?;
            // The rest of the reduce block gets put into a "body" block.
            write!(w, "{}_reduce_body:\n", self_name)?;
            // Extract the needed element from the used parallel vectors.
            self.vector_width.set(Some(width));
            for virt_reg in vectors_from_parallel.iter() {
                write!(w, "  %extract_v{} = extractelement ", virt_reg)?;
                self.emit_svalue(&SValue::VirtualRegister(*virt_reg), true, w)?;
                write!(w, ", i64 %linear_{}\n", block_idx)?;
            }
            self.vector_width.set(None);

            // Signal that the terminator needs to be a conditional branch to
            // close the loop.
            self.vector_reduce_cycle.set(true);
        }

        let vector_reduce_associative_vars = block
            .insts
            .iter()
            .enumerate()
            .filter_map(|(inst_idx, inst)| {
                inst.try_reduction_variable()
                    .map(|num| (inst_idx, block.virt_regs[inst_idx].0, num))
            })
            .filter(|(inst_idx, _, _)| block.schedules[&inst_idx].contains(&SSchedule::Associative))
            .map(|(_, virt_reg, num)| (virt_reg, num))
            .collect();

        self.vectors_from_parallel.replace(vectors_from_parallel);
        self.vector_reduce_associative_vars
            .replace(vector_reduce_associative_vars);

        Ok(())
    }

    /*
     * Reset the cells storing block specific context configuration.
     */
    pub fn reset_cells(&self) {
        self.vector_width.take();
        self.outside_def_used_in_vector.take();
        self.vectors_from_parallel.take();
        self.vector_reduce_associative_vars.take();
        self.vector_reduce_cycle.take();
    }
}

impl SBinaryOperator {
    fn get_llvm_op(&self, left_ty: &SType) -> &'static str {
        enum OpTy {
            Float,
            Unsigned,
            Signed,
        }

        let op_ty = if left_ty.is_float() {
            OpTy::Float
        } else if left_ty.is_unsigned() {
            OpTy::Unsigned
        } else {
            OpTy::Signed
        };

        match (self, op_ty) {
            (SBinaryOperator::Add, OpTy::Float) => "fadd",
            (SBinaryOperator::Add, _) => "add",
            (SBinaryOperator::Sub, OpTy::Float) => "fsub",
            (SBinaryOperator::Sub, _) => "sub",
            (SBinaryOperator::Mul, OpTy::Float) => "fmul",
            (SBinaryOperator::Mul, _) => "mul",
            (SBinaryOperator::Div, OpTy::Float) => "fdiv",
            (SBinaryOperator::Div, OpTy::Unsigned) => "udiv",
            (SBinaryOperator::Div, OpTy::Signed) => "sdiv",
            (SBinaryOperator::Rem, OpTy::Float) => "frem",
            (SBinaryOperator::Rem, OpTy::Unsigned) => "urem",
            (SBinaryOperator::Rem, OpTy::Signed) => "srem",
            (SBinaryOperator::LT, OpTy::Float) => "fcmp olt",
            (SBinaryOperator::LT, OpTy::Unsigned) => "icmp ult",
            (SBinaryOperator::LT, OpTy::Signed) => "icmp slt",
            (SBinaryOperator::LTE, OpTy::Float) => "fcmp ole",
            (SBinaryOperator::LTE, OpTy::Unsigned) => "icmp ule",
            (SBinaryOperator::LTE, OpTy::Signed) => "icmp sle",
            (SBinaryOperator::GT, OpTy::Float) => "fcmp ogt",
            (SBinaryOperator::GT, OpTy::Unsigned) => "icmp ugt",
            (SBinaryOperator::GT, OpTy::Signed) => "icmp sgt",
            (SBinaryOperator::GTE, OpTy::Float) => "fcmp oge",
            (SBinaryOperator::GTE, OpTy::Unsigned) => "icmp uge",
            (SBinaryOperator::GTE, OpTy::Signed) => "icmp sge",
            (SBinaryOperator::EQ, OpTy::Float) => "fcmp oeq",
            (SBinaryOperator::EQ, _) => "icmp eq",
            (SBinaryOperator::NE, OpTy::Float) => "fcmp one",
            (SBinaryOperator::NE, _) => "icmp ne",
            (SBinaryOperator::Or, _) => "or",
            (SBinaryOperator::And, _) => "and",
            (SBinaryOperator::Xor, _) => "xor",
            (SBinaryOperator::LSh, _) => "lsh",
            (SBinaryOperator::RSh, OpTy::Unsigned) => "lshr",
            (SBinaryOperator::RSh, _) => "ashr",
        }
    }
}
