extern crate serde;

extern crate hercules_ir;

use std::iter::once;

use self::serde::Deserialize;
use self::serde::Serialize;

use self::hercules_ir::*;

use crate::*;

/*
 * A manifest stores metadata about a Hercules function. This metadata is used
 * by the runtime to actually call a Hercules function.
 */
#[derive(Debug, Clone, Hash, Serialize, Deserialize)]
pub struct Manifest {
    // The signature of each Hercules function is represented in terms of
    // STypes, since this is the lowest level type representation that Hercules
    // constructs before reaching target-specific backends.
    pub param_types: Vec<(SType, ParameterKind)>,
    pub return_type: SType,

    // The dynamic constants (potentially) used in this Hercules function.
    pub dynamic_constants: Vec<DynamicConstant>,
    // The dimensions for array constants defined and used in this Hercules
    // function.
    pub array_constants: Vec<Box<[DynamicConstantID]>>,

    // The partitions that make up this Hercules function.
    pub partitions: Vec<PartitionManifest>,
}

#[derive(Debug, Clone, Hash, Serialize, Deserialize)]
pub struct PartitionManifest {
    // Each partition has one corresponding SFunction.
    pub name: SFunctionName,
    // Record the type and kind of each parameter.
    pub parameters: Vec<(SType, ParameterKind)>,
    // Record the type and kind of each return value.
    pub returns: Vec<(SType, ReturnKind)>,
    // Record the list of possible successors from this partition.
    pub successors: Vec<PartitionID>,
    // Device specific parts of the manifest. Represents details of calling
    // partition functions not present in the schedule IR type information
    // (since schedule IR is target independent).
    pub device: DeviceManifest,
}

#[derive(Debug, Clone, Hash, Serialize, Deserialize, PartialEq, Eq)]
pub enum ParameterKind {
    // A parameter corresponding to a parameter of the Hercules function.
    HerculesParameter(usize),
    // A parameter corresponding to some data defined in some other partition.
    DataInput(NodeID),
    // A parameter corresponding to a dynamic constant input to the Hercules
    // function.
    DynamicConstant(usize),
    // A parameter corresponding to an array constant used in the partition.
    ArrayConstant(ArrayID),
}

#[derive(Debug, Clone, Hash, Serialize, Deserialize)]
pub enum ReturnKind {
    // A return value corresponding to the return value of the Hercules
    // function.
    HerculesReturn,
    // A return value corresponding to some data used in some other partition.
    DataOutput(NodeID),
    // An integer specifying which partition should be executed next, if this
    // partition has multiple successors.
    NextPartition,
}

#[derive(Debug, Clone, Hash, Serialize, Deserialize)]
pub enum DeviceManifest {
    CPU {
        // If there's a top level fork-join that we parallel launch, specify,
        // for each thread dimension, how many tiles we want to spawn, and
        // the thread count. The thread count is a dynamic constant.
        parallel_launch: Box<[(usize, DynamicConstantID)]>,
    },
    GPU,
}

impl Manifest {
    pub fn all_visible_types(&self) -> impl Iterator<Item = SType> + '_ {
        self.param_types
            // Include the Hercules function parameter types.
            .iter()
            .map(|(ty, _)| ty.clone())
            // Include the Hercules function return type.
            .chain(once(self.return_type.clone()))
            // Include the partition parameter types.
            .chain(
                self.partitions
                    .iter()
                    .map(|partition| partition.parameters.iter().map(|(ty, _)| ty.clone()))
                    .flatten(),
            )
            // Include the partition return types.
            .chain(
                self.partitions
                    .iter()
                    .map(|partition| partition.returns.iter().map(|(ty, _)| ty.clone()))
                    .flatten(),
            )
            // Include the product types formed by the partition return types,
            // since multiple return values are returned inside a struct.
            .chain(self.partitions.iter().map(|partition| {
                SType::Product(partition.returns.iter().map(|(ty, _)| ty.clone()).collect())
            }))
    }
}

impl PartitionManifest {
    pub fn data_inputs(&self) -> impl Iterator<Item = (NodeID, &SType)> + '_ {
        self.parameters.iter().filter_map(|(stype, param_kind)| {
            if let ParameterKind::DataInput(id) = param_kind {
                Some((*id, stype))
            } else {
                None
            }
        })
    }

    pub fn data_outputs(&self) -> impl Iterator<Item = (NodeID, &SType)> + '_ {
        self.returns.iter().filter_map(|(stype, return_kind)| {
            if let ReturnKind::DataOutput(id) = return_kind {
                Some((*id, stype))
            } else {
                None
            }
        })
    }
}

impl DeviceManifest {
    pub fn default(device: Device) -> Self {
        match device {
            Device::CPU => DeviceManifest::CPU {
                parallel_launch: Box::new([]),
            },
            Device::GPU => todo!(),
        }
    }

    pub fn num_parallel_launch_dims(&self) -> usize {
        match self {
            DeviceManifest::CPU { parallel_launch } => parallel_launch.len(),
            DeviceManifest::GPU => 0,
        }
    }
}
