extern crate bitvec;
extern crate rand;

use std::collections::{HashMap, VecDeque};
use std::env::temp_dir;
use std::fmt::Write;
use std::fs::File;
use std::io::Write as _;
use std::process::Command;

use self::bitvec::prelude::*;

use self::rand::Rng;

use crate::*;

/*
 * Top level function to compute a dot graph for a schedule IR module, and
 * immediately render it using xdot.
 */
pub fn xdot_sched_module(module: &SModule) {
    let mut tmp_path = temp_dir();
    let mut rng = rand::thread_rng();
    let num: u64 = rng.gen();
    tmp_path.push(format!("sched_dot_{}.dot", num));
    let mut file = File::create(tmp_path.clone()).expect("PANIC: Unable to open output file.");
    let mut contents = String::new();
    write_dot(&module, &mut contents).expect("PANIC: Unable to generate output file contents.");
    file.write_all(contents.as_bytes())
        .expect("PANIC: Unable to write output file contents.");
    Command::new("xdot")
        .args([tmp_path])
        .output()
        .expect("PANIC: Couldn't execute xdot. Is xdot installed?");
}

/*
 * Top level function to write a schedule IR module out as a dot graph.
 */
pub fn write_dot<W: Write>(module: &SModule, w: &mut W) -> std::fmt::Result {
    write_digraph_header(w)?;

    for (function_name, function) in module.functions.iter() {
        // Schedule the SFunction to form a linear ordering of instructions.
        let virt_reg_to_inst_id = sched_virt_reg_to_inst_id(function);
        let dep_graph = sched_dependence_graph(function, &virt_reg_to_inst_id);
        let mut block_to_inst_list = (0..function.blocks.len())
            .map(|block_idx| (block_idx, vec![]))
            .collect::<HashMap<usize, Vec<(&SInst, usize, Option<&Vec<SSchedule>>)>>>();
        for (block_idx, block) in function.blocks.iter().enumerate() {
            let mut emitted = bitvec![u8, Lsb0; 0; block.insts.len()];
            let mut worklist = VecDeque::from((0..block.insts.len()).collect::<Vec<_>>());
            while let Some(inst_idx) = worklist.pop_front() {
                let inst_id = InstID::new(block_idx, inst_idx);
                let dependencies = &dep_graph[&inst_id];
                let all_uses_emitted = dependencies
                    .into_iter()
                    // Check that all used instructions in this block...
                    .filter(|inst_id| inst_id.idx_0() == block_idx)
                    // were already emitted.
                    .all(|inst_id| emitted[inst_id.idx_1()]);
                // Phis don't need to wait for all of their uses to be added.
                if block.insts[inst_idx].is_phi() || all_uses_emitted {
                    block_to_inst_list.get_mut(&block_idx).unwrap().push((
                        &block.insts[inst_idx],
                        block.virt_regs[inst_idx].0,
                        block.schedules.get(&inst_idx),
                    ));
                    emitted.set(inst_id.idx_1(), true);
                } else {
                    worklist.push_back(inst_idx);
                }
            }
        }

        // A SFunction is a subgraph.
        write_subgraph_header(function_name, w)?;

        // Each SBlock is a record node.
        for (block_idx, block) in function.blocks.iter().enumerate() {
            // Emit the instructions in scheduled order.
            write_block(function_name, block_idx, &block_to_inst_list[&block_idx], w)?;

            // Add control edges.
            for succ in block.successors().as_ref() {
                write_control_edge(function_name, block_idx, succ.idx(), w)?;
            }
        }

        write_graph_footer(w)?;
    }

    write_graph_footer(w)?;
    Ok(())
}

fn write_digraph_header<W: Write>(w: &mut W) -> std::fmt::Result {
    write!(w, "digraph \"Module\" {{\n")?;
    write!(w, "compound=true\n")?;
    Ok(())
}

fn write_subgraph_header<W: Write>(function_name: &SFunctionName, w: &mut W) -> std::fmt::Result {
    write!(w, "subgraph {} {{\n", function_name)?;
    write!(w, "label=\"{}\"\n", function_name)?;
    write!(w, "bgcolor=ivory4\n")?;
    write!(w, "cluster=true\n")?;
    Ok(())
}

fn write_graph_footer<W: Write>(w: &mut W) -> std::fmt::Result {
    write!(w, "}}\n")?;
    Ok(())
}

fn write_block<W: Write>(
    function_name: &SFunctionName,
    block_idx: usize,
    insts: &[(&SInst, usize, Option<&Vec<SSchedule>>)],
    w: &mut W,
) -> std::fmt::Result {
    write!(w, "{}_{} [label=\"{{", function_name, block_idx,)?;
    for token in insts.into_iter().map(|token| Some(token)).intersperse(None) {
        match token {
            Some((inst, virt_reg, schedules)) => {
                write!(w, "%{} = {}(", virt_reg, inst.upper_case_name())?;
                for token in sched_get_uses(inst).map(|u| Some(u)).intersperse(None) {
                    match token {
                        Some(SValue::VirtualRegister(use_virt_reg)) => {
                            write!(w, "%{}", use_virt_reg)?
                        }
                        Some(SValue::Constant(scons)) => write!(w, "{:?}", scons)?,
                        None => write!(w, ", ")?,
                    }
                }
                write!(w, ")")?;
                if let Some(schedules) = schedules
                    && !schedules.is_empty()
                {
                    write!(w, " [")?;
                    for token in schedules.into_iter().map(|s| Some(s)).intersperse(None) {
                        match token {
                            Some(schedule) => write!(w, "{:?}", schedule)?,
                            None => write!(w, ", ")?,
                        }
                    }
                    write!(w, "]")?;
                }
            }
            None => write!(w, " | ")?,
        }
    }
    write!(
        w,
        "}}\", shape = \"Mrecord\", style = \"filled\", fillcolor = \"lightblue\"];\n"
    )?;
    Ok(())
}

fn write_control_edge<W: Write>(
    function_name: &SFunctionName,
    src: usize,
    dst: usize,
    w: &mut W,
) -> std::fmt::Result {
    write!(
        w,
        "{}_{} -> {}_{} [color=\"black\"];\n",
        function_name, src, function_name, dst
    )?;
    Ok(())
}
